%{
#include "plasma.h"
%}

%token TOK_EOF TOK_EOL TOK_INDENT
%token <num> TOK_INT
%token <str> TOK_ID TOK_STRING
%right '#' '='
%left '|' '&' '^'
%left TOK_SHL TOK_SHR
%left '+' '-'
%left '*' '/'
%right '~' '@'

%type <term> opt_label opt_exprs expr

%union {
	term_t			term;
	long long		num;
	char			*str;
}

%parse-param {term_t p_label}
%parse-param {term_t p_body}

%%

line:		opt_label opt_exprs TOK_EOL	{ PL_unify(p_label, $1); PL_unify(p_body, $2); YYACCEPT; }
	|	TOK_EOL				{ PL_unify_nil(p_label); PL_unify_nil(p_body); YYACCEPT; }
	|	error TOK_EOL			{ yyerrok; YYACCEPT; }
	|	TOK_EOF				{ YYABORT; }
	;

opt_label:	TOK_ID				{ $$ = PL_new_term_ref(); PL_put_atom_chars($$, globalize($1, 1)); free($1); }
	|	TOK_INDENT			{ $$ = PL_new_term_ref(); PL_put_nil($$); }
	;

opt_exprs:	expr opt_exprs			{ $$ = PL_new_term_ref(); PL_cons_list($$, $1, $2); }
	|	expr ',' opt_exprs		{ $$ = PL_new_term_ref(); PL_cons_list($$, $1, $3); }
	|	/* empty */			{ $$ = PL_new_term_ref(); PL_put_nil($$); }
	;

expr:
		TOK_INT				{ $$ = PL_new_term_ref(); PL_put_int64($$, $1); }
	|	TOK_STRING			{ $$ = PL_new_term_ref(); PL_put_string_chars($$, $1); }
	|	TOK_ID				{ $$ = PL_new_term_ref(); PL_put_atom_chars($$, globalize($1, 0)); free($1); }
	|	'#' expr			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("imm"), 1), $2); }
	|	'=' expr			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("pool"), 1), $2); }
	|	'@' expr			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("hub"), 1), $2); }
	|	'~' expr			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("not"), 1), $2); }
	|	'(' '-' expr ')'		{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("neg"), 1), $3); }
	|	expr '+' expr			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("add"), 2), $1, $3); }
	|	expr '-' expr			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("sub"), 2), $1, $3); }
	|	expr '|' expr			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("or" ), 2), $1, $3); }
	|	expr '&' expr			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("and"), 2), $1, $3); }
	|	expr '^' expr			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("xor"), 2), $1, $3); }
	|	expr '*' expr			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("mul"), 2), $1, $3); }
	|	expr '/' expr			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("div"), 2), $1, $3); }
	|	expr TOK_SHL expr		{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("shl"), 2), $1, $3); }
	|	expr TOK_SHR expr		{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("shr"), 2), $1, $3); }
	|	'[' expr ']'			{ $$ = PL_new_term_ref(); PL_cons_functor($$, PL_new_functor(PL_new_atom("repeat"), 1), $2); }
	|	'(' expr ')'			{ $$ = $2; }
	;
